from django.db import models
from edc_utils import get_utcnow

from edc_vitals.model_mixins import (
    BloodPressureModelMixin,
    SimpleBloodPressureModelMixin,
    WeightHeightBmiModelMixin,
)
from edc_vitals.models import HeartRateField, RespiratoryRateField, TemperatureField


class BloodPressure(BloodPressureModelMixin, models.Model):
    pass


class HeartRate(models.Model):
    heart_rate = HeartRateField(
        null=True,
        blank=True,
    )


class RespiratoryRate(models.Model):
    respiratory_rate = RespiratoryRateField(
        null=True,
        blank=True,
    )


class SimpleBloodPressure(SimpleBloodPressureModelMixin, models.Model):
    pass


class Temperature(models.Model):
    temperature = TemperatureField(
        null=True,
        blank=True,
    )


class WeightHeightBmi(WeightHeightBmiModelMixin, models.Model):
    report_datetime = models.DateTimeField(default=get_utcnow)

    dob = models.DateField(null=True)



from dateutil.relativedelta import relativedelta
from django import forms
from django.test import TestCase, tag
from edc_constants.constants import BLACK, MALE
from edc_form_validators import FormValidator
from edc_utils import get_utcnow
from edc_utils.round_up import round_half_away_from_zero

from edc_vitals.calculators import BMI, CalculatorError, calculate_bmi
from edc_vitals.form_validators import BmiFormValidatorMixin


class TestCalculators(TestCase):
    @tag("1")
    def test_bmi_calculator(self):
        dob = get_utcnow() - relativedelta(years=25)
        self.assertRaises(CalculatorError, BMI, weight_kg=56, height_cm=None)
        try:
            calculate_bmi(weight_kg=56, height_cm=None, dob=dob)
        except CalculatorError:
            self.fail("CalculatorError unexpectedly raised ")

        for func in [BMI, calculate_bmi]:
            with self.subTest(func=func):
                self.assertRaises(
                    CalculatorError,
                    func,
                    weight_kg=56,
                    height_cm=1.50,
                    dob=dob,
                    report_datetime=get_utcnow(),
                )
                try:
                    bmi = func(
                        weight_kg=56, height_cm=150, dob=dob, report_datetime=get_utcnow()
                    )
                except CalculatorError as e:
                    self.fail(f"CalculatorError unexpectedly raises. Got {e}")
                else:
                    self.assertEqual(round_half_away_from_zero(bmi.value, 2), 24.89)

    @tag("1")
    def test_bmi_form_validator(self):
        data = dict(
            gender=MALE,
            ethnicity=BLACK,
            age_in_years=30,
        )

        class BmiFormValidator(BmiFormValidatorMixin, FormValidator):
            pass

        # not enough data
        form_validator = BmiFormValidator(cleaned_data=data)
        bmi = form_validator.validate_bmi()
        self.assertIsNone(bmi)

        # calculates
        data.update(
            weight=56,
            height=150,
            dob=get_utcnow() - relativedelta(years=30),
            report_datetime=get_utcnow(),
        )
        form_validator = BmiFormValidator(cleaned_data=data)
        bmi = form_validator.validate_bmi()
        self.assertEqual(bmi.value, 24.8889)

        # calculation error
        data.update(
            weight=56,
            height=1.5,
            dob=get_utcnow() - relativedelta(years=25),
            report_datetime=get_utcnow(),
        )
        form_validator = BmiFormValidator(cleaned_data=data)
        self.assertRaises(forms.ValidationError, form_validator.validate_bmi)



from django.core.exceptions import ValidationError
from django.test import TestCase

from ..models import HeartRate


class TestTemperature(TestCase):
    def test_simple_ok(self):
        obj = HeartRate(heart_rate=45)
        obj.save()
        self.assertEqual(obj.heart_rate, 45)

    def test_heart_rate_lt_30_raises(self):
        for low_hr in [0, 1, 28, 29]:
            with self.subTest(heart_rate=low_hr):
                model = HeartRate.objects.create(heart_rate=low_hr)
                with self.assertRaises(ValidationError) as cm:
                    model.full_clean()
                self.assertIn("heart_rate", cm.exception.error_dict)
                self.assertIn(
                    "Ensure this value is greater than or equal to 30.",
                    str(cm.exception.error_dict.get("heart_rate")),
                )

    def test_30_lt_heart_rate_lt_200_ok(self):
        for hr in [30, 31, 45, 60, 120, 199, 200]:
            with self.subTest(heart_rate=hr):
                model = HeartRate.objects.create(heart_rate=hr)
                try:
                    model.full_clean()
                except ValidationError as e:
                    self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_heart_rate_gt_200_raises(self):
        for high_hr in [201, 202, 300]:
            with self.subTest(heart_rate=high_hr):
                model = HeartRate.objects.create(heart_rate=high_hr)
                with self.assertRaises(ValidationError) as cm:
                    model.full_clean()
                self.assertIn("heart_rate", cm.exception.error_dict)
                self.assertIn(
                    "Ensure this value is less than or equal to 200.",
                    str(cm.exception.error_dict.get("heart_rate")),
                )



from django import forms
from django.core.exceptions import ValidationError
from django.test import TestCase
from edc_constants.constants import YES

from edc_vitals.form_validators import BloodPressureFormValidatorMixin
from edc_vitals.utils import has_severe_htn

from ..models import BloodPressure, SimpleBloodPressure


class TestBloodPressure(TestCase):
    def test_allows_none(self):
        obj = BloodPressure()
        obj.save()

    def test_calculates_avg_only_if_fields_provided(self):
        obj = BloodPressure(sys_blood_pressure_one=100)
        obj.save()
        obj.refresh_from_db()
        self.assertIsNone(obj.sys_blood_pressure_avg, obj.dia_blood_pressure_avg)

        obj = BloodPressure(sys_blood_pressure_one=100, dia_blood_pressure_one=100)
        obj.save()
        obj.refresh_from_db()
        self.assertIsNone(obj.sys_blood_pressure_avg, obj.dia_blood_pressure_avg)

        obj = BloodPressure(
            sys_blood_pressure_one=100,
            dia_blood_pressure_one=100,
            sys_blood_pressure_two=100,
            dia_blood_pressure_two=100,
        )
        obj.save()
        obj.refresh_from_db()
        self.assertEqual(obj.sys_blood_pressure_avg, 100)
        self.assertEqual(obj.dia_blood_pressure_avg, 100)

    def test_calculates_avg(self):
        obj = BloodPressure(
            sys_blood_pressure_one=110,
            dia_blood_pressure_one=80,
            sys_blood_pressure_two=128,
            dia_blood_pressure_two=91,
        )
        obj.save()
        obj.refresh_from_db()
        self.assertEqual(obj.sys_blood_pressure_avg, 119)
        self.assertEqual(obj.dia_blood_pressure_avg, 85)

    def test_simple_allows_none(self):
        obj = SimpleBloodPressure()
        obj.save()

    def test_simple_ok(self):
        obj = SimpleBloodPressure(
            sys_blood_pressure=110,
            dia_blood_pressure=80,
        )
        obj.save()

    def test_has_severe_htn(self):
        self.assertIsNone(has_severe_htn())
        self.assertIsNone(has_severe_htn(sys=120))
        self.assertFalse(has_severe_htn(sys=120, dia=80))
        self.assertFalse(has_severe_htn(sys=179, dia=109))
        self.assertTrue(has_severe_htn(sys=180, dia=109))
        self.assertTrue(has_severe_htn(sys=179, dia=110))
        self.assertTrue(has_severe_htn(sys=180, dia=110))


class TestBloodPressureFormValidatorMixin(TestCase):
    def test_bp_suggest_severe_htn_form_validator(self):
        cleaned_data = dict(
            sys_blood_pressure_one=None,
            sys_blood_pressure_two=None,
            dia_blood_pressure_one=None,
            dia_blood_pressure_two=None,
        )
        form_validator = BloodPressureFormValidatorMixin()
        self.assertIsNone(
            form_validator.raise_on_avg_blood_pressure_suggests_severe_htn(**cleaned_data)
        )
        cleaned_data.update(sys_blood_pressure_one=180, dia_blood_pressure_one=None)
        self.assertIsNone(
            form_validator.raise_on_avg_blood_pressure_suggests_severe_htn(**cleaned_data)
        )
        cleaned_data.update(sys_blood_pressure_one=180, dia_blood_pressure_one=120)
        self.assertIsNone(
            form_validator.raise_on_avg_blood_pressure_suggests_severe_htn(**cleaned_data)
        )
        cleaned_data.update(
            sys_blood_pressure_one=180,
            dia_blood_pressure_one=120,
            sys_blood_pressure_two=191,
            dia_blood_pressure_two=125,
        )
        self.assertRaises(
            forms.ValidationError,
            form_validator.raise_on_avg_blood_pressure_suggests_severe_htn,
            **cleaned_data,
        )

        cleaned_data.update(severe_htn=YES)
        try:
            form_validator.raise_on_avg_blood_pressure_suggests_severe_htn(**cleaned_data)
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_sys_lt_dia_bp_raises(self):
        cleaned_data = dict(
            sys_blood_pressure=80,
            dia_blood_pressure=120,
        )
        form_validator = BloodPressureFormValidatorMixin()

        with self.assertRaises(ValidationError) as cm:
            form_validator.raise_on_systolic_lt_diastolic_bp(**cleaned_data)
        self.assertIn("dia_blood_pressure", cm.exception.error_dict)
        self.assertIn(
            "Invalid. Diastolic must be less than systolic.",
            str(cm.exception.error_dict.get("dia_blood_pressure")),
        )

        cleaned_data = dict(
            custom_sys_blood_pressure_fld=80,
            custom_dia_blood_pressure_fld=120,
        )
        form_validator = BloodPressureFormValidatorMixin()

        with self.assertRaises(ValidationError) as cm:
            form_validator.raise_on_systolic_lt_diastolic_bp(
                sys_field="custom_sys_blood_pressure_fld",
                dia_field="custom_dia_blood_pressure_fld",
                **cleaned_data,
            )
        self.assertIn("custom_dia_blood_pressure_fld", cm.exception.error_dict)
        self.assertIn(
            "Invalid. Diastolic must be less than systolic.",
            str(cm.exception.error_dict.get("custom_dia_blood_pressure_fld")),
        )

    def test_sys_gte_dia_bp_ok(self):
        for sys, dia in [(120, 80), (180, 110), (100, 99), (90, 90)]:
            with self.subTest(sys=sys, dia=dia):
                cleaned_data = dict(
                    sys_blood_pressure=sys,
                    dia_blood_pressure=dia,
                )
                form_validator = BloodPressureFormValidatorMixin()
                try:
                    form_validator.raise_on_systolic_lt_diastolic_bp(**cleaned_data)
                except ValidationError as e:
                    self.fail(f"ValidationError unexpectedly raised. Got {e}")

                cleaned_data = dict(
                    custom_sys_blood_pressure_fld=sys,
                    custom_dia_blood_pressure_fld=dia,
                )
                try:
                    form_validator.raise_on_systolic_lt_diastolic_bp(
                        sys_field="custom_sys_blood_pressure_fld",
                        dia_field="custom_dia_blood_pressure_fld",
                        **cleaned_data,
                    )
                except ValidationError as e:
                    self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_raise_on_systolic_lt_diastolic_bp_with_none_ok(self):
        for sys, dia in [(None, None), (None, 80), (120, None)]:
            with self.subTest(sys=sys, dia=dia):
                cleaned_data = dict(
                    sys_blood_pressure=sys,
                    dia_blood_pressure=dia,
                )
                form_validator = BloodPressureFormValidatorMixin()
                try:
                    form_validator.raise_on_systolic_lt_diastolic_bp(**cleaned_data)
                except ValidationError as e:
                    self.fail(f"ValidationError unexpectedly raised. Got {e}")

                cleaned_data = dict(
                    custom_sys_blood_pressure_fld=sys,
                    custom_dia_blood_pressure_fld=dia,
                )
                try:
                    form_validator.raise_on_systolic_lt_diastolic_bp(
                        sys_field="custom_sys_blood_pressure_fld",
                        dia_field="custom_dia_blood_pressure_fld",
                        **cleaned_data,
                    )
                except ValidationError as e:
                    self.fail(f"ValidationError unexpectedly raised. Got {e}")



from django.core.exceptions import ValidationError
from django.test import TestCase

from edc_vitals.utils import (
    get_g3_fever_lower,
    get_g4_fever_lower,
    has_g3_fever,
    has_g4_fever,
)

from ..models import Temperature


class TestTemperature(TestCase):
    def test_simple_ok(self):
        obj = Temperature(temperature=37.5)
        obj.save()
        self.assertEqual(obj.temperature, 37.5)

    def test_temperature_lt_30_raises(self):
        for low_temp in [28.0, 29, 29.9]:
            with self.subTest(temperature=low_temp):
                model = Temperature.objects.create(temperature=low_temp)
                with self.assertRaises(ValidationError) as cm:
                    model.full_clean()
                self.assertIn("temperature", cm.exception.error_dict)
                self.assertIn(
                    "Ensure this value is greater than or equal to 30.",
                    str(cm.exception.error_dict.get("temperature")),
                )

    def test_30_lt_temperature_lt_45_ok(self):
        for temperature in [30, 31, 37.5, 39.3, 40, 45]:
            with self.subTest(temperature=temperature):
                model = Temperature.objects.create(temperature=temperature)
                try:
                    model.full_clean()
                except ValidationError as e:
                    self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_temperature_gt_45_raises(self):
        for high_temp in [45.1, 46, 50.0]:
            with self.subTest(temperature=high_temp):
                model = Temperature.objects.create(temperature=high_temp)
                with self.assertRaises(ValidationError) as cm:
                    model.full_clean()
                self.assertIn("temperature", cm.exception.error_dict)
                self.assertIn(
                    "Ensure this value is less than or equal to 45.",
                    str(cm.exception.error_dict.get("temperature")),
                )

    def test_has_fever_passed_none_returns_none(self):
        self.assertIsNone(has_g3_fever())
        self.assertIsNone(has_g4_fever())

    def test_has_g3_fever_returns_false_if_temp_lt_g3(self):
        for temperature in [36, 37, 37.5, 38, 38.6, 39.2]:
            with self.subTest(temperature=temperature):
                self.assertFalse(has_g3_fever(temperature=temperature))

    def test_has_g3_fever_returns_true_if_temp_g3(self):
        for temperature in [39.3, 39.5, 39.9]:
            with self.subTest(temperature=temperature):
                self.assertTrue(has_g3_fever(temperature=temperature))

    def test_has_g3_fever_returns_false_if_temp_gt_g3(self):
        for temperature in [40, 40.1, 41, 45, 50]:
            with self.subTest(temperature=temperature):
                self.assertFalse(has_g3_fever(temperature=temperature))

    def test_has_g4_fever_returns_false_if_temp_lt_g4(self):
        for temperature in [36, 39.2, 39.3, 39.5, 39.9]:
            with self.subTest(temperature=temperature):
                self.assertFalse(has_g4_fever(temperature=temperature))

    def test_has_g4_fever_returns_true_if_temp_g4(self):
        for temperature in [40, 40.1, 41, 45, 50]:
            with self.subTest(temperature=temperature):
                self.assertTrue(has_g4_fever(temperature=temperature))

    def test_get_g3_fever_lower(self):
        self.assertEqual(get_g3_fever_lower(), 39.3)

    def test_get_g4_fever_lower(self):
        self.assertEqual(get_g4_fever_lower(), 40.0)



from django.core.exceptions import ValidationError
from django.test import TestCase

from ..models import RespiratoryRate


class TestTemperature(TestCase):
    def test_simple_ok(self):
        obj = RespiratoryRate(respiratory_rate=20)
        obj.save()
        self.assertEqual(obj.respiratory_rate, 20)

    def test_respiratory_rate_lt_6_raises(self):
        for low_rr in [0, 1, 5]:
            with self.subTest(respiratory_rate=low_rr):
                model = RespiratoryRate.objects.create(respiratory_rate=low_rr)
                with self.assertRaises(ValidationError) as cm:
                    model.full_clean()
                self.assertIn("respiratory_rate", cm.exception.error_dict)
                self.assertIn(
                    "Ensure this value is greater than or equal to 6.",
                    str(cm.exception.error_dict.get("respiratory_rate")),
                )

    def test_6_lt_respiratory_rate_lt_50_ok(self):
        for rr in [6, 7, 10, 20, 49, 50]:
            with self.subTest(respiratory_rate=rr):
                model = RespiratoryRate.objects.create(respiratory_rate=rr)
                try:
                    model.full_clean()
                except ValidationError as e:
                    self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_respiratory_rate_gt_50_raises(self):
        for high_rr in [51, 52, 60, 100]:
            with self.subTest(respiratory_rate=high_rr):
                model = RespiratoryRate.objects.create(respiratory_rate=high_rr)
                with self.assertRaises(ValidationError) as cm:
                    model.full_clean()
                self.assertIn("respiratory_rate", cm.exception.error_dict)
                self.assertIn(
                    "Ensure this value is less than or equal to 50.",
                    str(cm.exception.error_dict.get("respiratory_rate")),
                )
